@@ -22,6 +22,7 @@
 #ifndef __CONNMAN_PROVIDER_H
 #define __CONNMAN_PROVIDER_H
 
+#include <glib.h>
 #include <connman/types.h>
 
 #ifdef __cplusplus
@@ -56,6 +57,7 @@ enum connman_provider_error {
 };
 
 struct connman_provider;
+struct connman_ipaddress;
 
 struct connman_provider *connman_provider_ref(struct connman_provider *provider);
 void connman_provider_unref(struct connman_provider *provider);
@@ -76,22 +78,19 @@ int connman_provider_get_index(struct connman_provider *provider);
 
 void connman_provider_set_data(struct connman_provider *provider, void *data);
 void *connman_provider_get_data(struct connman_provider *provider);
-
-void connman_provider_set_gateway(struct connman_provider *provider,
-							const char *gateway);
-void connman_provider_set_address(struct connman_provider *provider,
-							const char *address);
-void connman_provider_set_netmask(struct connman_provider *provider,
-							const char *netmask);
-void connman_provider_set_dns(struct connman_provider *provider,
-							const char *dns);
-void connman_provider_set_domain(struct connman_provider *provider,
-							const char *domain);
-
+int connman_provider_set_ipaddress(struct connman_provider *provider,
+					struct connman_ipaddress *ipaddress);
+int connman_provider_set_pac(struct connman_provider *provider,
+				const char *pac);
+int connman_provider_set_domain(struct connman_provider *provider,
+				const char *domain);
+int connman_provider_set_nameservers(struct connman_provider *provider,
+					const char *nameservers);
 int connman_provider_append_route(struct connman_provider *provider,
 					const char *key, const char *value);
 
 const char *connman_provider_get_driver_name(struct connman_provider *provider);
+const char *connman_provider_get_save_group(struct connman_provider *provider);
 
 struct connman_provider_driver {
 	const char *name;
@@ -100,6 +99,7 @@ struct connman_provider_driver {
 	int (*remove) (struct connman_provider *provider);
 	int (*connect) (struct connman_provider *provider);
 	int (*disconnect) (struct connman_provider *provider);
+	int (*save) (struct connman_provider *provider, GKeyFile *keyfile);
 };
 
 int connman_provider_driver_register(struct connman_provider_driver *driver);
