@@ -87,7 +87,7 @@ int __connman_agent_unregister(const char *sender, const char *path)
 
 struct request_input_reply {
 	struct connman_service *service;
-	passphrase_cb_t callback;
+	authentication_cb_t callback;
 	void *user_data;
 };
 
@@ -95,6 +95,7 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 {
 	struct request_input_reply *passphrase_reply = user_data;
 	connman_bool_t wps = FALSE;
+	char *identity = NULL;
 	char *passphrase = NULL;
 	char *wpspin = NULL;
 	char *key;
@@ -114,13 +115,22 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 			break;
 
 		dbus_message_iter_get_basic(&entry, &key);
-		if (g_str_equal(key, "Passphrase")) {
+
+		if (g_str_equal(key, "Identity")) {
+			dbus_message_iter_next(&entry);
+			if (dbus_message_iter_get_arg_type(&entry)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_recurse(&entry, &value);
+			dbus_message_iter_get_basic(&value, &identity);
+
+		} else if (g_str_equal(key, "Passphrase")) {
 			dbus_message_iter_next(&entry);
 			if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_VARIANT)
 				break;
 			dbus_message_iter_recurse(&entry, &value);
 			dbus_message_iter_get_basic(&value, &passphrase);
-			break;
+
 		} else if (g_str_equal(key, "WPS")) {
 			wps = TRUE;
 
@@ -154,7 +164,7 @@ static void request_input_passphrase_reply(DBusPendingCall *call, void *user_dat
 	}
 
 done:
-	passphrase_reply->callback(passphrase_reply->service,
+	passphrase_reply->callback(passphrase_reply->service, identity,
 				passphrase, passphrase_reply->user_data);
 	connman_service_unref(passphrase_reply->service);
 	dbus_message_unref(reply);
@@ -181,11 +191,24 @@ static void request_input_append_alternates(DBusMessageIter *iter,
 	g_strfreev(alternates);
 }
 
+static void request_input_append_identity(DBusMessageIter *iter,
+							void *user_data)
+{
+	char *str = "string";
+
+	connman_dbus_dict_append_basic(iter, "Type",
+				DBUS_TYPE_STRING, &str);
+	str = "Mandatory";
+	connman_dbus_dict_append_basic(iter, "Requirement",
+				DBUS_TYPE_STRING, &str);
+}
+
 static void request_input_append_passphrase(DBusMessageIter *iter,
 							void *user_data)
 {
 	struct connman_service *service = user_data;
 	char *value;
+	const char *phase2;
 
 	switch (__connman_service_get_security(service)) {
 	case CONNMAN_SERVICE_SECURITY_WEP:
@@ -193,6 +216,17 @@ static void request_input_append_passphrase(DBusMessageIter *iter,
 		break;
 	case CONNMAN_SERVICE_SECURITY_PSK:
 		value = "psk";
+		break;
+	case CONNMAN_SERVICE_SECURITY_8021X:
+		phase2 = __connman_service_get_phase2(service);
+
+		if (phase2 != NULL && (
+				g_str_has_suffix(phase2, "GTC") == TRUE ||
+				g_str_has_suffix(phase2, "OTP") == TRUE))
+			value = "response";
+		else
+			value = "passphrase";
+
 		break;
 	default:
 		value = "string";
@@ -223,8 +257,72 @@ static void request_input_append_wps(DBusMessageIter *iter, void *user_data)
 				DBUS_TYPE_STRING, &str);
 }
 
-int __connman_agent_request_input(struct connman_service *service,
-				passphrase_cb_t callback, void *user_data)
+static void request_input_append_password(DBusMessageIter *iter,
+							void *user_data)
+{
+	char *str = "passphrase";
+
+	connman_dbus_dict_append_basic(iter, "Type",
+				DBUS_TYPE_STRING, &str);
+	str = "Mandatory";
+	connman_dbus_dict_append_basic(iter, "Requirement",
+				DBUS_TYPE_STRING, &str);
+}
+
+static void request_input_login_reply(DBusPendingCall *call, void *user_data)
+{
+	struct request_input_reply *username_password_reply = user_data;
+	char *username = NULL;
+	char *password = NULL;
+	char *key;
+	DBusMessageIter iter, dict;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto done;
+
+	dbus_message_iter_init(reply, &iter);
+	dbus_message_iter_recurse(&iter, &dict);
+	while (dbus_message_iter_get_arg_type(&dict) == DBUS_TYPE_DICT_ENTRY) {
+		DBusMessageIter entry, value;
+
+		dbus_message_iter_recurse(&dict, &entry);
+		if (dbus_message_iter_get_arg_type(&entry) != DBUS_TYPE_STRING)
+			break;
+
+		dbus_message_iter_get_basic(&entry, &key);
+
+		if (g_str_equal(key, "Username")) {
+			dbus_message_iter_next(&entry);
+			if (dbus_message_iter_get_arg_type(&entry)
+							!= DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_recurse(&entry, &value);
+			dbus_message_iter_get_basic(&value, &username);
+
+		} else if (g_str_equal(key, "Password")) {
+			dbus_message_iter_next(&entry);
+			if (dbus_message_iter_get_arg_type(&entry) !=
+							DBUS_TYPE_VARIANT)
+				break;
+			dbus_message_iter_recurse(&entry, &value);
+			dbus_message_iter_get_basic(&value, &password);
+		}
+
+		dbus_message_iter_next(&dict);
+	}
+
+done:
+	username_password_reply->callback(username_password_reply->service,
+					username, password,
+					username_password_reply->user_data);
+	connman_service_unref(username_password_reply->service);
+	dbus_message_unref(reply);
+	g_free(username_password_reply);
+}
+
+int __connman_agent_request_passphrase_input(struct connman_service *service,
+				authentication_cb_t callback, void *user_data)
 {
 	DBusMessage *message;
 	const char *path;
@@ -249,6 +347,13 @@ int __connman_agent_request_input(struct connman_service *service,
 				DBUS_TYPE_OBJECT_PATH, &path);
 
 	connman_dbus_dict_open(&iter, &dict);
+
+	if (__connman_service_get_security(service) ==
+			CONNMAN_SERVICE_SECURITY_8021X) {
+		connman_dbus_dict_append_dict(&dict, "Identity",
+					request_input_append_identity, service);
+	}
+
 	connman_dbus_dict_append_dict(&dict, "Passphrase",
 				request_input_append_passphrase, service);
 
@@ -290,6 +395,72 @@ int __connman_agent_request_input(struct connman_service *service,
 	return -EIO;
 }
 
+int __connman_agent_request_login_input(struct connman_service *service,
+				authentication_cb_t callback, void *user_data)
+{
+	DBusMessage *message;
+	const char *path;
+	DBusMessageIter iter;
+	DBusMessageIter dict;
+	DBusPendingCall *call;
+	struct request_input_reply *username_password_reply;
+
+	if (service == NULL || agent_path == NULL || callback == NULL)
+		return -ESRCH;
+
+	message = dbus_message_new_method_call(agent_sender, agent_path,
+					CONNMAN_AGENT_INTERFACE,
+					"RequestInput");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_iter_init_append(message, &iter);
+
+	path = __connman_service_get_path(service);
+	dbus_message_iter_append_basic(&iter,
+				DBUS_TYPE_OBJECT_PATH, &path);
+
+	connman_dbus_dict_open(&iter, &dict);
+
+	connman_dbus_dict_append_dict(&dict, "Username",
+				request_input_append_identity, service);
+
+	connman_dbus_dict_append_dict(&dict, "Password",
+				request_input_append_password, service);
+
+	connman_dbus_dict_close(&iter, &dict);
+
+	username_password_reply = g_try_new0(struct request_input_reply, 1);
+	if (username_password_reply == NULL) {
+		dbus_message_unref(message);
+		return -ENOMEM;
+	}
+
+	if (dbus_connection_send_with_reply(connection, message,
+							&call, -1) == FALSE) {
+		dbus_message_unref(message);
+		g_free(username_password_reply);
+		return -ESRCH;
+	}
+
+	if (call == NULL) {
+		dbus_message_unref(message);
+		g_free(username_password_reply);
+		return -ESRCH;
+	}
+
+	username_password_reply->service = connman_service_ref(service);
+	username_password_reply->callback = callback;
+	username_password_reply->user_data = user_data;
+
+	dbus_pending_call_set_notify(call, request_input_login_reply,
+						username_password_reply, NULL);
+
+	dbus_message_unref(message);
+
+	return -EIO;
+}
+
 struct report_error_data {
 	struct connman_service *service;
 	report_error_cb_t callback;
