@@ -25,6 +25,7 @@
 #include <config.h>
 #endif
 
+#define _GNU_SOURCE
 #include <stdio.h>
 #include <errno.h>
 #include <unistd.h>
@@ -88,13 +89,15 @@ int __connman_inet_modify_address(int cmd, int flags,
 	struct in_addr ipv4_addr, ipv4_dest, ipv4_bcast;
 	int sk, err;
 
-	DBG("");
+	DBG("cmd %#x flags %#x index %d family %d address %s peer %s "
+		"prefixlen %hhu broadcast %s", cmd, flags, index, family,
+		address, peer, prefixlen, broadcast);
 
 	if (address == NULL)
-		return -1;
+		return -EINVAL;
 
 	if (family != AF_INET && family != AF_INET6)
-		return -1;
+		return -EINVAL;
 
 	memset(&request, 0, sizeof(request));
 
@@ -148,9 +151,9 @@ int __connman_inet_modify_address(int cmd, int flags,
 			return err;
 	}
 
-	sk = socket(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
+	sk = socket(AF_NETLINK, SOCK_DGRAM | SOCK_CLOEXEC, NETLINK_ROUTE);
 	if (sk < 0)
-		return -1;
+		return -errno;
 
 	memset(&nl_addr, 0, sizeof(nl_addr));
 	nl_addr.nl_family = AF_NETLINK;
@@ -175,7 +178,7 @@ int connman_inet_ifindex(const char *name)
 	if (name == NULL)
 		return -1;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -200,7 +203,7 @@ char *connman_inet_ifname(int index)
 	if (index < 0)
 		return NULL;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return NULL;
 
@@ -222,7 +225,7 @@ short int connman_inet_ifflags(int index)
 	struct ifreq ifr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -errno;
 
@@ -252,7 +255,7 @@ int connman_inet_ifup(int index)
 	struct ifreq ifr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -errno;
 
@@ -294,7 +297,7 @@ int connman_inet_ifdown(int index)
 	struct ifreq ifr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -errno;
 
@@ -339,7 +342,7 @@ static char *index2addr(int index)
 	if (index < 0)
 		return NULL;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return NULL;
 
@@ -382,7 +385,7 @@ static char *index2ident(int index, const char *prefix)
 	if (index < 0)
 		return NULL;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return NULL;
 
@@ -426,7 +429,7 @@ connman_bool_t connman_inet_is_cfg80211(int index)
 	struct ifreq ifr;
 	int sk;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return FALSE;
 
@@ -462,7 +465,7 @@ struct connman_device *connman_inet_create_device(int index)
 	if (devname == NULL)
 		return NULL;
 
-	if (__connman_element_device_isfiltered(devname) == TRUE) {
+	if (__connman_device_isfiltered(devname) == TRUE) {
 		connman_info("Ignoring interface %s (filtered)", devname);
 		free(devname);
 		return NULL;
@@ -541,6 +544,7 @@ struct in6_ifreq {
 int connman_inet_set_ipv6_address(int index,
 		struct connman_ipaddress *ipaddress)
 {
+	int err;
 	unsigned char prefix_len;
 	const char *address;
 
@@ -552,11 +556,12 @@ int connman_inet_set_ipv6_address(int index,
 
 	DBG("index %d address %s prefix_len %d", index, address, prefix_len);
 
-	if ((__connman_inet_modify_address(RTM_NEWADDR,
-			NLM_F_REPLACE | NLM_F_ACK, index, AF_INET6,
-				address, NULL, prefix_len, NULL)) < 0) {
-		connman_error("Set IPv6 address error");
-		return -1;
+	err = __connman_inet_modify_address(RTM_NEWADDR,
+				NLM_F_REPLACE | NLM_F_ACK, index, AF_INET6,
+				address, NULL, prefix_len, NULL);
+	if (err < 0) {
+		connman_error("%s: %s", __func__, strerror(-err));
+		return err;
 	}
 
 	return 0;
@@ -564,6 +569,7 @@ int connman_inet_set_ipv6_address(int index,
 
 int connman_inet_set_address(int index, struct connman_ipaddress *ipaddress)
 {
+	int err;
 	unsigned char prefix_len;
 	const char *address, *broadcast, *peer;
 
@@ -577,11 +583,12 @@ int connman_inet_set_address(int index, struct connman_ipaddress *ipaddress)
 
 	DBG("index %d address %s prefix_len %d", index, address, prefix_len);
 
-	if ((__connman_inet_modify_address(RTM_NEWADDR,
-			NLM_F_REPLACE | NLM_F_ACK, index, AF_INET,
-				address, peer, prefix_len, broadcast)) < 0) {
-		DBG("address setting failed");
-		return -1;
+	err = __connman_inet_modify_address(RTM_NEWADDR,
+				NLM_F_REPLACE | NLM_F_ACK, index, AF_INET,
+				address, peer, prefix_len, broadcast);
+	if (err < 0) {
+		connman_error("%s: %s", __func__, strerror(-err));
+		return err;
 	}
 
 	return 0;
@@ -590,12 +597,15 @@ int connman_inet_set_address(int index, struct connman_ipaddress *ipaddress)
 int connman_inet_clear_ipv6_address(int index, const char *address,
 							int prefix_len)
 {
+	int err;
+
 	DBG("index %d address %s prefix_len %d", index, address, prefix_len);
 
-	if ((__connman_inet_modify_address(RTM_DELADDR, 0, index, AF_INET6,
-					address, NULL, prefix_len, NULL)) < 0) {
-		connman_error("Clear IPv6 address error");
-		return -1;
+	err = __connman_inet_modify_address(RTM_DELADDR, 0, index, AF_INET6,
+				address, NULL, prefix_len, NULL);
+	if (err < 0) {
+		connman_error("%s: %s", __func__, strerror(-err));
+		return err;
 	}
 
 	return 0;
@@ -603,6 +613,7 @@ int connman_inet_clear_ipv6_address(int index, const char *address,
 
 int connman_inet_clear_address(int index, struct connman_ipaddress *ipaddress)
 {
+	int err;
 	unsigned char prefix_len;
 	const char *address, *broadcast, *peer;
 
@@ -613,10 +624,11 @@ int connman_inet_clear_address(int index, struct connman_ipaddress *ipaddress)
 
 	DBG("index %d address %s prefix_len %d", index, address, prefix_len);
 
-	if ((__connman_inet_modify_address(RTM_DELADDR, 0, index, AF_INET,
-				address, peer, prefix_len, broadcast)) < 0) {
-		DBG("address removal failed");
-		return -1;
+	err = __connman_inet_modify_address(RTM_DELADDR, 0, index, AF_INET,
+				address, peer, prefix_len, broadcast);
+	if (err < 0) {
+		connman_error("%s: %s", __func__, strerror(-err));
+		return err;
 	}
 
 	return 0;
@@ -642,7 +654,7 @@ int connman_inet_add_network_route(int index, const char *host,
 	struct sockaddr_in addr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -704,7 +716,7 @@ int connman_inet_del_network_route(int index, const char *host)
 	struct sockaddr_in addr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -762,7 +774,7 @@ int connman_inet_del_ipv6_network_route(int index, const char *host,
 	rt.rtmsg_metric = 1;
 	rt.rtmsg_ifindex = index;
 
-	sk = socket(AF_INET6, SOCK_DGRAM, 0);
+	sk = socket(AF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0) {
 		err = -1;
 		goto out;
@@ -772,7 +784,8 @@ int connman_inet_del_ipv6_network_route(int index, const char *host,
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Del IPv6 host route error");
+		connman_error("Del IPv6 host route error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -784,7 +797,7 @@ int connman_inet_del_ipv6_host_route(int index, const char *host)
 
 int connman_inet_add_ipv6_network_route(int index, const char *host,
 					const char *gateway,
-						unsigned char prefix_len)
+					unsigned char prefix_len)
 {
 	struct in6_rtmsg rt;
 	int sk, err;
@@ -812,7 +825,7 @@ int connman_inet_add_ipv6_network_route(int index, const char *host,
 	rt.rtmsg_metric = 1;
 	rt.rtmsg_ifindex = index;
 
-	sk = socket(AF_INET6, SOCK_DGRAM, 0);
+	sk = socket(AF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0) {
 		err = -1;
 		goto out;
@@ -822,7 +835,8 @@ int connman_inet_add_ipv6_network_route(int index, const char *host,
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Set IPv6 host route error");
+		connman_error("Set IPv6 host route error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -854,7 +868,7 @@ int connman_inet_set_ipv6_gateway_address(int index, const char *gateway)
 	rt.rtmsg_dst_len = 0;
 	rt.rtmsg_ifindex = index;
 
-	sk = socket(AF_INET6, SOCK_DGRAM, 0);
+	sk = socket(AF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0) {
 		err = -1;
 		goto out;
@@ -864,7 +878,8 @@ int connman_inet_set_ipv6_gateway_address(int index, const char *gateway)
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Set default IPv6 gateway error");
+		connman_error("Set default IPv6 gateway error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -890,7 +905,7 @@ int connman_inet_clear_ipv6_gateway_address(int index, const char *gateway)
 	rt.rtmsg_dst_len = 0;
 	rt.rtmsg_ifindex = index;
 
-	sk = socket(AF_INET6, SOCK_DGRAM, 0);
+	sk = socket(AF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0) {
 		err = -1;
 		goto out;
@@ -900,7 +915,8 @@ int connman_inet_clear_ipv6_gateway_address(int index, const char *gateway)
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Clear default IPv6 gateway error");
+		connman_error("Clear default IPv6 gateway error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -912,7 +928,7 @@ int connman_inet_set_gateway_address(int index, const char *gateway)
 	struct sockaddr_in addr;
 	int sk, err;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -963,7 +979,7 @@ int connman_inet_set_gateway_interface(int index)
 
 	DBG("");
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -999,6 +1015,52 @@ int connman_inet_set_gateway_interface(int index)
 	return err;
 }
 
+int connman_inet_set_ipv6_gateway_interface(int index)
+{
+	struct ifreq ifr;
+	struct rtentry rt;
+	struct sockaddr_in6 addr;
+	const struct in6_addr any = IN6ADDR_ANY_INIT;
+	int sk, err;
+
+	DBG("");
+
+	sk = socket(PF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
+	if (sk < 0)
+		return -1;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	DBG("ifname %s", ifr.ifr_name);
+
+	memset(&rt, 0, sizeof(rt));
+	rt.rt_flags = RTF_UP;
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin6_family = AF_INET6;
+	addr.sin6_addr = any;
+
+	memcpy(&rt.rt_genmask, &addr, sizeof(rt.rt_genmask));
+	memcpy(&rt.rt_dst, &addr, sizeof(rt.rt_dst));
+	memcpy(&rt.rt_gateway, &addr, sizeof(rt.rt_gateway));
+
+	rt.rt_dev = ifr.ifr_name;
+
+	err = ioctl(sk, SIOCADDRT, &rt);
+	if (err < 0)
+		connman_error("Setting default interface route failed (%s)",
+							strerror(errno));
+	close(sk);
+
+	return err;
+}
+
 int connman_inet_clear_gateway_address(int index, const char *gateway)
 {
 	struct ifreq ifr;
@@ -1008,7 +1070,7 @@ int connman_inet_clear_gateway_address(int index, const char *gateway)
 
 	DBG("");
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -1059,7 +1121,7 @@ int connman_inet_clear_gateway_interface(int index)
 
 	DBG("");
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return -1;
 
@@ -1095,6 +1157,52 @@ int connman_inet_clear_gateway_interface(int index)
 	return err;
 }
 
+int connman_inet_clear_ipv6_gateway_interface(int index)
+{
+	struct ifreq ifr;
+	struct rtentry rt;
+	struct sockaddr_in6 addr;
+	const struct in6_addr any = IN6ADDR_ANY_INIT;
+	int sk, err;
+
+	DBG("");
+
+	sk = socket(PF_INET6, SOCK_DGRAM | SOCK_CLOEXEC, 0);
+	if (sk < 0)
+		return -1;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	DBG("ifname %s", ifr.ifr_name);
+
+	memset(&rt, 0, sizeof(rt));
+	rt.rt_flags = RTF_UP;
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin6_family = AF_INET6;
+	addr.sin6_addr = any;
+
+	memcpy(&rt.rt_genmask, &addr, sizeof(rt.rt_genmask));
+	memcpy(&rt.rt_dst, &addr, sizeof(rt.rt_dst));
+	memcpy(&rt.rt_gateway, &addr, sizeof(rt.rt_gateway));
+
+	rt.rt_dev = ifr.ifr_name;
+
+	err = ioctl(sk, SIOCDELRT, &rt);
+	if (err < 0)
+		connman_error("Removing default interface route failed (%s)",
+							strerror(errno));
+	close(sk);
+
+	return err;
+}
+
 connman_bool_t connman_inet_compare_subnet(int index, const char *host)
 {
 	struct ifreq ifr;
@@ -1112,7 +1220,7 @@ connman_bool_t connman_inet_compare_subnet(int index, const char *host)
 		return -1;
 	host_addr = _host_addr.s_addr;
 
-	sk = socket(PF_INET, SOCK_DGRAM, 0);
+	sk = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return FALSE;
 
@@ -1136,6 +1244,9 @@ connman_bool_t connman_inet_compare_subnet(int index, const char *host)
 		close(sk);
 		return FALSE;
 	}
+
+	close(sk);
+
 	addr = (struct sockaddr_in *)&ifr.ifr_addr;
 	if_addr = addr->sin_addr.s_addr;
 
@@ -1150,7 +1261,7 @@ int connman_inet_remove_from_bridge(int index, const char *bridge)
 	if (bridge == NULL)
 		return -EINVAL;
 
-	sk = socket(AF_INET, SOCK_STREAM, 0);
+	sk = socket(AF_INET, SOCK_STREAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return sk;
 
@@ -1179,7 +1290,7 @@ int connman_inet_add_to_bridge(int index, const char *bridge)
 	if (bridge == NULL)
 		return -EINVAL;
 
-	sk = socket(AF_INET, SOCK_STREAM, 0);
+	sk = socket(AF_INET, SOCK_STREAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return sk;
 
@@ -1205,7 +1316,7 @@ int connman_inet_set_mtu(int index, int mtu)
 	struct ifreq ifr;
 	int sk, err;
 
-	sk = socket(AF_INET, SOCK_DGRAM, 0);
+	sk = socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return sk;
 
@@ -1232,7 +1343,7 @@ int connman_inet_setup_tunnel(char *tunnel, int mtu)
 	if (tunnel == NULL)
 		return -EINVAL;
 
-	sk = socket(AF_INET, SOCK_DGRAM, 0);
+	sk = socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (sk < 0)
 		return sk;
 
@@ -1272,7 +1383,7 @@ int connman_inet_create_tunnel(char **iface)
 	struct ifreq ifr;
 	int i, fd;
 
-	fd = open("/dev/net/tun", O_RDWR);
+	fd = open("/dev/net/tun", O_RDWR | O_CLOEXEC);
 	if (fd < 0) {
 		i = -errno;
 		connman_error("Failed to open /dev/net/tun: %s",
@@ -1306,6 +1417,7 @@ struct rs_cb_data {
 	__connman_inet_rs_cb_t callback;
 	struct sockaddr_in6 addr;
 	guint rs_timeout;
+	guint watch_id;
 	void *user_data;
 };
 
@@ -1319,12 +1431,6 @@ static const struct in6_addr in6addr_all_nodes_mc = IN6ADDR_ALL_NODES_MC_INIT;
 static const struct in6_addr in6addr_all_routers_mc =
 						IN6ADDR_ALL_ROUTERS_MC_INIT;
 
-/* from netinet/in.h */
-struct in6_pktinfo {
-	struct in6_addr ipi6_addr;  /* src/dst IPv6 address */
-	unsigned int ipi6_ifindex;  /* send/recv interface index */
-};
-
 static void rs_cleanup(struct rs_cb_data *data)
 {
 	g_io_channel_shutdown(data->channel, TRUE, NULL);
@@ -1334,6 +1440,9 @@ static void rs_cleanup(struct rs_cb_data *data)
 	if (data->rs_timeout > 0)
 		g_source_remove(data->rs_timeout);
 
+	if (data->watch_id > 0)
+		g_source_remove(data->watch_id);
+
 	g_free(data);
 }
 
@@ -1380,6 +1489,7 @@ static int icmpv6_recv(int fd, gpointer user_data)
 	len = recvmsg(fd, &mhdr, 0);
 	if (len < 0) {
 		data->callback(NULL, data->user_data);
+		rs_cleanup(data);
 		return -errno;
 	}
 
@@ -1468,7 +1578,7 @@ static int ndisc_send_unspec(int type, int oif, const struct in6_addr *dest)
 
 	DBG("");
 
-	fd = socket(AF_INET6, SOCK_RAW, IPPROTO_RAW);
+	fd = socket(AF_INET6, SOCK_RAW | SOCK_CLOEXEC, IPPROTO_RAW);
 	if (fd < 0)
 		return -errno;
 
@@ -1582,7 +1692,7 @@ int __connman_inet_ipv6_send_rs(int index, int timeout,
 	data->user_data = user_data;
 	data->rs_timeout = g_timeout_add_seconds(timeout, rs_timeout_cb, data);
 
-	sk = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6);
+	sk = socket(AF_INET6, SOCK_RAW | SOCK_CLOEXEC, IPPROTO_ICMPV6);
 	if (sk < 0)
 		return -errno;
 
@@ -1602,7 +1712,7 @@ int __connman_inet_ipv6_send_rs(int index, int timeout,
 	g_io_channel_set_encoding(data->channel, NULL, NULL);
 	g_io_channel_set_buffered(data->channel, FALSE);
 
-	g_io_add_watch(data->channel,
+	data->watch_id = g_io_add_watch(data->channel,
 			G_IO_IN | G_IO_NVAL | G_IO_HUP | G_IO_ERR,
 			icmpv6_event, data);
 
