@@ -23,10 +23,15 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
+
 #include <gdbus.h>
 
 #include "connman.h"
 
+connman_bool_t connman_state_idle;
+DBusMessage *session_mode_pending = NULL;
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -45,11 +50,6 @@ static DBusMessage *get_properties(DBusConnection *conn,
 
 	connman_dbus_dict_open(&array, &dict);
 
-	str = __connman_profile_active_path();
-	if (str != NULL)
-		connman_dbus_dict_append_basic(&dict, "ActiveProfile",
-						DBUS_TYPE_OBJECT_PATH, &str);
-
 	connman_dbus_dict_append_array(&dict, "Services",
 			DBUS_TYPE_OBJECT_PATH, __connman_service_list, NULL);
 	connman_dbus_dict_append_array(&dict, "Technologies",
@@ -59,7 +59,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_basic(&dict, "State",
 						DBUS_TYPE_STRING, &str);
 
-	offlinemode = __connman_profile_get_offlinemode();
+	offlinemode = __connman_technology_get_offlinemode();
 	connman_dbus_dict_append_basic(&dict, "OfflineMode",
 					DBUS_TYPE_BOOLEAN, &offlinemode);
 
@@ -116,15 +116,7 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&value, &offlinemode);
 
-		__connman_profile_set_offlinemode(offlinemode, TRUE);
-
-		__connman_profile_save_default();
-	} else if (g_str_equal(name, "ActiveProfile") == TRUE) {
-		const char *str;
-
-		dbus_message_iter_get_basic(&value, &str);
-
-		return __connman_error_not_supported(msg);
+		__connman_technology_set_offlinemode(offlinemode);
 	} else if (g_str_equal(name, "SessionMode") == TRUE) {
 		connman_bool_t sessionmode;
 
@@ -133,7 +125,16 @@ static DBusMessage *set_property(DBusConnection *conn,
 
 		dbus_message_iter_get_basic(&value, &sessionmode);
 
+		if (session_mode_pending != NULL)
+			return __connman_error_in_progress(msg);
+
 		__connman_session_set_mode(sessionmode);
+
+		if (sessionmode == TRUE && connman_state_idle == FALSE) {
+			session_mode_pending = msg;
+			return NULL;
+		}
+
 	} else
 		return __connman_error_invalid_property(msg);
 
@@ -192,7 +193,7 @@ static DBusMessage *request_scan(DBusConnection *conn,
 	else
 		return __connman_error_invalid_arguments(msg);
 
-	err = __connman_element_request_scan(type);
+	err = __connman_device_request_scan(type);
 	if (err < 0) {
 		if (err == -EINPROGRESS) {
 			connman_error("Invalid return code from scan");
@@ -207,63 +208,34 @@ static DBusMessage *request_scan(DBusConnection *conn,
 
 static DBusConnection *connection = NULL;
 
-static enum connman_service_type technology_type;
-static connman_bool_t technology_enabled;
-static DBusMessage *technology_pending = NULL;
-static guint technology_timeout = 0;
-
-static void technology_reply(int error)
+static void session_mode_notify(void)
 {
-	DBG("");
+	DBusMessage *reply;
 
-	if (technology_timeout > 0) {
-		g_source_remove(technology_timeout);
-		technology_timeout = 0;
-	}
+	reply = g_dbus_create_reply(session_mode_pending, DBUS_TYPE_INVALID);
+	g_dbus_send_message(connection, reply);
 
-	if (technology_pending != NULL) {
-		if (error > 0) {
-			DBusMessage *reply;
-
-			reply = __connman_error_failed(technology_pending,
-								error);
-			if (reply != NULL)
-				g_dbus_send_message(connection, reply);
-		} else
-			g_dbus_send_reply(connection, technology_pending,
-							DBUS_TYPE_INVALID);
-
-		dbus_message_unref(technology_pending);
-		technology_pending = NULL;
-	}
-
-	technology_type = CONNMAN_SERVICE_TYPE_UNKNOWN;
+	dbus_message_unref(session_mode_pending);
+	session_mode_pending = NULL;
 }
 
-static gboolean technology_abort(gpointer user_data)
+static void idle_state(connman_bool_t idle)
 {
-	DBG("");
 
-	technology_timeout = 0;
+	DBG("idle %d", idle);
 
-	technology_reply(ETIMEDOUT);
+	connman_state_idle = idle;
 
-	return FALSE;
-}
+	if (connman_state_idle == FALSE || session_mode_pending == NULL)
+		return;
 
-static void technology_notify(enum connman_service_type type,
-						connman_bool_t enabled)
-{
-	DBG("type %d enabled %d", type, enabled);
-
-	if (type == technology_type && enabled == technology_enabled)
-		technology_reply(0);
+	session_mode_notify();
 }
 
 static struct connman_notifier technology_notifier = {
 	.name		= "manager",
 	.priority	= CONNMAN_NOTIFIER_PRIORITY_HIGH,
-	.service_enabled= technology_notify,
+	.idle_state	= idle_state,
 };
 
 static DBusMessage *enable_technology(DBusConnection *conn,
@@ -271,13 +243,9 @@ static DBusMessage *enable_technology(DBusConnection *conn,
 {
 	enum connman_service_type type;
 	const char *str;
-	int err;
 
 	DBG("conn %p", conn);
 
-	if (technology_pending != NULL)
-		return __connman_error_in_progress(msg);
-
 	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
 							DBUS_TYPE_INVALID);
 
@@ -300,16 +268,7 @@ static DBusMessage *enable_technology(DBusConnection *conn,
 	if (__connman_notifier_is_enabled(type) == TRUE)
 		return __connman_error_already_enabled(msg);
 
-	technology_type = type;
-	technology_enabled = TRUE;
-	technology_pending = dbus_message_ref(msg);
-
-	err = __connman_element_enable_technology(type);
-	if (err < 0 && err != -EINPROGRESS)
-		technology_reply(-err);
-	else
-		technology_timeout = g_timeout_add_seconds(15,
-						technology_abort, NULL);
+	 __connman_technology_enable(type, msg);
 
 	return NULL;
 }
@@ -319,13 +278,9 @@ static DBusMessage *disable_technology(DBusConnection *conn,
 {
 	enum connman_service_type type;
 	const char *str;
-	int err;
 
 	DBG("conn %p", conn);
 
-	if (technology_pending != NULL)
-		return __connman_error_in_progress(msg);
-
 	dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &str,
 							DBUS_TYPE_INVALID);
 
@@ -348,16 +303,7 @@ static DBusMessage *disable_technology(DBusConnection *conn,
 	if (__connman_notifier_is_enabled(type) == FALSE)
 		return __connman_error_already_disabled(msg);
 
-	technology_type = type;
-	technology_enabled = FALSE;
-	technology_pending = dbus_message_ref(msg);
-
-	err = __connman_element_disable_technology(type);
-	if (err < 0 && err != -EINPROGRESS)
-		technology_reply(-err);
-	else
-		technology_timeout = g_timeout_add_seconds(10,
-						technology_abort, NULL);
+	__connman_technology_disable(type, msg);
 
 	return NULL;
 }
@@ -611,14 +557,15 @@ static DBusMessage *request_private_network(DBusConnection *conn,
 static DBusMessage *release_private_network(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	const char *sender;
+	const char *path;
 	int err;
 
 	DBG("conn %p", conn);
 
-	sender = dbus_message_get_sender(msg);
+	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+							DBUS_TYPE_INVALID);
 
-	err = __connman_private_network_release(sender);
+	err = __connman_private_network_release(path);
 	if (err < 0)
 		return __connman_error_failed(msg, -err);
 
@@ -627,7 +574,8 @@ static DBusMessage *release_private_network(DBusConnection *conn,
 
 static GDBusMethodTable manager_methods[] = {
 	{ "GetProperties",     "",      "a{sv}", get_properties     },
-	{ "SetProperty",       "sv",    "",      set_property       },
+	{ "SetProperty",       "sv",    "",      set_property,
+						G_DBUS_METHOD_FLAG_ASYNC },
 	{ "GetState",          "",      "s",     get_state          },
 	{ "RemoveProvider",    "o",     "",      remove_provider    },
 	{ "RequestScan",       "s",     "",      request_scan       },
@@ -649,10 +597,10 @@ static GDBusMethodTable manager_methods[] = {
 	{ "UnregisterCounter", "o",     "",      unregister_counter },
 	{ "CreateSession",     "a{sv}o", "o",    create_session     },
 	{ "DestroySession",    "o",     "",      destroy_session    },
-	{ "RequestPrivateNetwork",    "",     "ha{sv}",
+	{ "RequestPrivateNetwork",    "",     "oa{sv}h",
 						request_private_network,
 						G_DBUS_METHOD_FLAG_ASYNC },
-	{ "ReleasePrivateNetwork",    "",     "",
+	{ "ReleasePrivateNetwork",    "o",    "",
 						release_private_network },
 	{ },
 };
@@ -679,6 +627,8 @@ int __connman_manager_init(void)
 					manager_methods,
 					manager_signals, NULL, NULL, NULL);
 
+	connman_state_idle = TRUE;
+
 	return 0;
 }
 
@@ -686,11 +636,11 @@ void __connman_manager_cleanup(void)
 {
 	DBG("");
 
-	connman_notifier_unregister(&technology_notifier);
-
 	if (connection == NULL)
 		return;
 
+	connman_notifier_unregister(&technology_notifier);
+
 	g_dbus_unregister_interface(connection, CONNMAN_MANAGER_PATH,
 						CONNMAN_MANAGER_INTERFACE);
 
