@@ -30,6 +30,7 @@
 static DBusConnection *connection = NULL;
 
 static GSList *notifier_list = NULL;
+static GHashTable *service_hash = NULL;
 
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
@@ -72,21 +73,22 @@ void connman_notifier_unregister(struct connman_notifier *notifier)
 
 #define MAX_TECHNOLOGIES 10
 
-static volatile gint registered[MAX_TECHNOLOGIES];
-static volatile gint enabled[MAX_TECHNOLOGIES];
-static volatile gint connected[MAX_TECHNOLOGIES];
+static volatile int registered[MAX_TECHNOLOGIES];
+static volatile int enabled[MAX_TECHNOLOGIES];
+static volatile int connected[MAX_TECHNOLOGIES];
 
 void __connman_notifier_list_registered(DBusMessageIter *iter, void *user_data)
 {
 	int i;
 
+	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
 		const char *type = __connman_service_type2string(i);
 
 		if (type == NULL)
 			continue;
 
-		if (g_atomic_int_get(&registered[i]) > 0)
+		if (registered[i] > 0)
 			dbus_message_iter_append_basic(iter,
 						DBUS_TYPE_STRING, &type);
 	}
@@ -96,13 +98,14 @@ void __connman_notifier_list_enabled(DBusMessageIter *iter, void *user_data)
 {
 	int i;
 
+	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
 		const char *type = __connman_service_type2string(i);
 
 		if (type == NULL)
 			continue;
 
-		if (g_atomic_int_get(&enabled[i]) > 0)
+		if (enabled[i] > 0)
 			dbus_message_iter_append_basic(iter,
 						DBUS_TYPE_STRING, &type);
 	}
@@ -112,13 +115,14 @@ void __connman_notifier_list_connected(DBusMessageIter *iter, void *user_data)
 {
 	int i;
 
+	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
 		const char *type = __connman_service_type2string(i);
 
 		if (type == NULL)
 			continue;
 
-		if (g_atomic_int_get(&connected[i]) > 0)
+		if (connected[i] > 0)
 			dbus_message_iter_append_basic(iter,
 						DBUS_TYPE_STRING, &type);
 	}
@@ -157,8 +161,9 @@ unsigned int __connman_notifier_count_connected(void)
 {
 	unsigned int i, count = 0;
 
+	__sync_synchronize();
 	for (i = 0; i < MAX_TECHNOLOGIES; i++) {
-		if (g_atomic_int_get(&connected[i]) > 0)
+		if (connected[i] > 0)
 			count++;
 	}
 
@@ -175,7 +180,7 @@ const char *__connman_notifier_get_state(void)
 	return "offline";
 }
 
-static void state_changed(void)
+static void state_changed(connman_bool_t connected)
 {
 	unsigned int count = __connman_notifier_count_connected();
 	char *state = "offline";
@@ -184,8 +189,12 @@ static void state_changed(void)
 	if (count > 1)
 		return;
 
-	if (count > 0)
+	if (count == 1) {
+		if (connected == FALSE)
+			return;
+
 		state = "online";
+	}
 
 	connman_dbus_property_changed_basic(CONNMAN_MANAGER_PATH,
 				CONNMAN_MANAGER_INTERFACE, "State",
@@ -211,7 +220,7 @@ static void technology_connected(enum connman_service_type type,
 		CONNMAN_MANAGER_INTERFACE, "ConnectedTechnologies",
 		DBUS_TYPE_STRING, __connman_notifier_list_connected, NULL);
 
-	state_changed();
+	state_changed(connected);
 }
 
 void __connman_notifier_register(enum connman_service_type type)
@@ -233,7 +242,7 @@ void __connman_notifier_register(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_exchange_and_add(&registered[type], 1) == 0)
+	if (__sync_fetch_and_add(&registered[type], 1) == 0)
 		technology_registered(type, TRUE);
 }
 
@@ -241,7 +250,8 @@ void __connman_notifier_unregister(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
-	if (g_atomic_int_get(&registered[type]) == 0) {
+	__sync_synchronize();
+	if (registered[type] == 0) {
 		connman_error("notifier unregister underflow");
 		return;
 	}
@@ -261,8 +271,10 @@ void __connman_notifier_unregister(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_dec_and_test(&registered[type]) == TRUE)
-		technology_registered(type, FALSE);
+	if (__sync_fetch_and_sub(&registered[type], 1) != 1)
+		return;
+
+	technology_registered(type, FALSE);
 }
 
 void __connman_notifier_enable(enum connman_service_type type)
@@ -284,7 +296,7 @@ void __connman_notifier_enable(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_exchange_and_add(&enabled[type], 1) == 0)
+	if (__sync_fetch_and_add(&enabled[type], 1) == 0)
 		technology_enabled(type, TRUE);
 }
 
@@ -292,7 +304,8 @@ void __connman_notifier_disable(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
-	if (g_atomic_int_get(&enabled[type]) == 0) {
+	__sync_synchronize();
+	if (enabled[type] == 0) {
 		connman_error("notifier disable underflow");
 		return;
 	}
@@ -312,8 +325,10 @@ void __connman_notifier_disable(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_dec_and_test(&enabled[type]) == TRUE)
-		technology_enabled(type, FALSE);
+	if (__sync_fetch_and_sub(&enabled[type], 1) != 1)
+		return;
+
+	technology_enabled(type, FALSE);
 }
 
 void __connman_notifier_connect(enum connman_service_type type)
@@ -335,7 +350,7 @@ void __connman_notifier_connect(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_exchange_and_add(&connected[type], 1) == 0)
+	if (__sync_fetch_and_add(&connected[type], 1) == 0)
 		technology_connected(type, TRUE);
 }
 
@@ -343,7 +358,8 @@ void __connman_notifier_disconnect(enum connman_service_type type)
 {
 	DBG("type %d", type);
 
-	if (g_atomic_int_get(&connected[type]) == 0) {
+	__sync_synchronize();
+	if (connected[type] == 0) {
 		connman_error("notifier disconnect underflow");
 		return;
 	}
@@ -363,8 +379,10 @@ void __connman_notifier_disconnect(enum connman_service_type type)
 		break;
 	}
 
-	if (g_atomic_int_dec_and_test(&connected[type]) == TRUE)
-		technology_connected(type, FALSE);
+	if (__sync_fetch_and_sub(&connected[type], 1) != 1)
+		return;
+
+	technology_connected(type, FALSE);
 }
 
 static void technology_default(enum connman_service_type type)
@@ -400,7 +418,8 @@ void __connman_notifier_default_changed(struct connman_service *service)
 	}
 }
 
-void __connman_notifier_service_add(struct connman_service *service)
+void __connman_notifier_service_add(struct connman_service *service,
+					const char *name)
 {
 	GSList *list;
 
@@ -408,7 +427,7 @@ void __connman_notifier_service_add(struct connman_service *service)
 		struct connman_notifier *notifier = list->data;
 
 		if (notifier->service_add)
-			notifier->service_add(service);
+			notifier->service_add(service, name);
 	}
 }
 
@@ -416,6 +435,17 @@ void __connman_notifier_service_remove(struct connman_service *service)
 {
 	GSList *list;
 
+	if (g_hash_table_lookup(service_hash, service) != NULL) {
+		/*
+		 * This is a tempory check for consistency. It can be
+		 * removed when there are no reports for the following
+		 * error message.
+		 */
+		connman_error("Service state machine inconsistency detected.");
+
+		g_hash_table_remove(service_hash, service);
+	}
+
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
 
@@ -451,8 +481,6 @@ void __connman_notifier_offlinemode(connman_bool_t enabled)
 
 	DBG("enabled %d", enabled);
 
-	__connman_profile_changed(FALSE);
-
 	offlinemode_changed(enabled);
 
 	for (list = notifier_list; list; list = list->next) {
@@ -463,10 +491,26 @@ void __connman_notifier_offlinemode(connman_bool_t enabled)
 	}
 }
 
+static void notify_idle_state(connman_bool_t idle)
+{
+	GSList *list;
+
+	DBG("idle %d", idle);
+
+	for (list = notifier_list; list; list = list->next) {
+		struct connman_notifier *notifier = list->data;
+
+		if (notifier->idle_state)
+			notifier->idle_state(idle);
+	}
+}
+
 void __connman_notifier_service_state_changed(struct connman_service *service,
 					enum connman_service_state state)
 {
 	GSList *list;
+	unsigned int old_size;
+	connman_bool_t found;
 
 	for (list = notifier_list; list; list = list->next) {
 		struct connman_notifier *notifier = list->data;
@@ -474,6 +518,36 @@ void __connman_notifier_service_state_changed(struct connman_service *service,
 		if (notifier->service_state_changed)
 			notifier->service_state_changed(service, state);
 	}
+
+	old_size = g_hash_table_size(service_hash);
+	found = g_hash_table_lookup(service_hash, service) != NULL;
+
+	switch (state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_IDLE:
+		if (found == FALSE)
+			break;
+
+		g_hash_table_remove(service_hash, service);
+		if (old_size == 1)
+			notify_idle_state(TRUE);
+
+		break;
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_ONLINE:
+		if (found == TRUE)
+			break;
+
+		g_hash_table_insert(service_hash, service, service);
+		if (old_size == 0)
+			notify_idle_state(FALSE);
+
+		break;
+	}
 }
 
 void __connman_notifier_ipconfig_changed(struct connman_service *service,
@@ -516,7 +590,8 @@ connman_bool_t __connman_notifier_is_registered(enum connman_service_type type)
 	if (technology_supported(type) == FALSE)
 		return FALSE;
 
-	if (g_atomic_int_get(&registered[type]) > 0)
+	__sync_synchronize();
+	if (registered[type] > 0)
 		return TRUE;
 
 	return FALSE;
@@ -529,7 +604,8 @@ connman_bool_t __connman_notifier_is_enabled(enum connman_service_type type)
 	if (technology_supported(type) == FALSE)
 		return FALSE;
 
-	if (g_atomic_int_get(&enabled[type]) > 0)
+	__sync_synchronize();
+	if (enabled[type] > 0)
 		return TRUE;
 
 	return FALSE;
@@ -541,6 +617,10 @@ int __connman_notifier_init(void)
 
 	connection = connman_dbus_get_connection();
 
+	service_hash = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+						NULL, NULL);
+
+
 	return 0;
 }
 
@@ -548,5 +628,8 @@ void __connman_notifier_cleanup(void)
 {
 	DBG("");
 
+	g_hash_table_destroy(service_hash);
+	service_hash = NULL;
+
 	dbus_connection_unref(connection);
 }
